"""
analysis.py

This archive.py defines to calculated POES from material balance equation.

The logic is structure using Classes and methods.

libraries:
    - pandas
    - matplotlib
    - pydantic
    - pandera
    - scipy
    - typing
"""

import pandas as pd
import pandera as pa
import numpy as np
from matplotlib.font_manager import FontProperties
from matplotlib.ticker import FuncFormatter
from pandera.typing import Series
from matplotlib import pyplot as plt
from pydantic import BaseModel
from scipy import stats
from scipy.interpolate import UnivariateSpline
from typing import Union, Optional
from pytank.constants.constants import (
    OIL_FVF_COL,
    GAS_FVF_COL,
    RS_COL,
    PRESSURE_COL,
    OIL_CUM_COL,
    GAS_CUM_COL,
    WATER_CUM_COL,
    DATE_COL,
    WELL_COL,
    WATER_FVF_COL,
    RS_W_COL,
    TANK_COL,
    UW_COL,
    PRESSURE_PVT_COL,
    OIL_CUM_TANK,
    WATER_CUM_TANK,
    GAS_CUM_TANK,
    OIL_EXP,
    RES_EXP,
    WE,
    OIL_RATE_COL,
    WATER_RATE_COL,
)
from pytank.functions.utilities import interp_dates_row
from pytank.functions.material_balance import (
    underground_withdrawal,
    pressure_vol_avg,
    ho_terms_equation,
    calculated_pressure_fetkovich,
    calculate_pressure_with_carter_tracy,
)
from pytank.tank import Tank
from pytank.aquifer_model import Fetkovich, CarterTracy


class _PressSchema(pa.DataFrameModel):
    """
    Private Class to validate data of df_press_int method in Tank Class
    """

    PRESSURE_DATUM: Series[float] = pa.Field(nullable=False)
    WELL_BORE: Series[str] = pa.Field(nullable=False)
    START_DATETIME: Series[pd.Timestamp] = pa.Field(nullable=False)
    Bo: Series[float] = pa.Field(nullable=False)
    Bg: Series[float] = pa.Field(nullable=True)
    GOR: Series[float] = pa.Field(nullable=False)
    Bw: Series[float] = pa.Field(nullable=False)
    RS_bw: Series[float] = pa.Field(nullable=False)
    Tank: Series[str] = pa.Field(nullable=False)


class _ProdSchema(pa.DataFrameModel):
    """
    Private Class to validate data of df_prod_int method in Tank Class
    """

    OIL_CUM: Series[float] = pa.Field(nullable=False)
    WATER_CUM: Series[float] = pa.Field(nullable=False)
    GAS_CUM: Series[float] = pa.Field(nullable=False)
    LIQ_CUM: Series[float] = pa.Field(nullable=False)
    WELL_BORE: Series[str] = pa.Field(nullable=False)
    START_DATETIME: Series[pd.Timestamp] = pa.Field(nullable=False)
    Tank: Series[str] = pa.Field(nullable=False)


class _DFMbalSchema(pa.DataFrameModel):
    Tank: Series[str] = pa.Field(nullable=False)
    START_DATETIME: Series[pd.Timestamp] = pa.Field(nullable=False)
    PRESSURE_DATUM: Series[float] = pa.Field(nullable=False)
    OIL_CUM_TANK: Series[float] = pa.Field(nullable=False)
    WATER_CUM_TANK: Series[float] = pa.Field(nullable=False)
    GAS_CUM_TANK: Series[float] = pa.Field(nullable=False)
    Bo: Series[float] = pa.Field(nullable=False)
    Bg: Series[float] = pa.Field(nullable=True)
    RS_bw: Series[float] = pa.Field(nullable=False)
    Bw: Series[float] = pa.Field(nullable=False)
    GOR: Series[float] = pa.Field(nullable=False)
    Time_Step: Series[float] = pa.Field(nullable=False)
    UW: Series[float] = pa.Field(nullable=False)
    Eo: Series[float] = pa.Field(nullable=False)
    Efw: Series[float] = pa.Field(nullable=False)


class Analysis(BaseModel):
    """
    Class to calculate POES graphically and analytically.
    It also allows to project graphics:
        - DataFrame with necessary information to material balance
        - Campbell
        - Exploratory data analysis (EDA) through graphics

    Parameters
    ----------
    tank_class : Tank
        Instance of the Tank class.
    freq : str
        Frequency of data for the material balance equation.
    position : str
        Position of the frequency of the date.
    smooth : bool, optional
        Determines if the user wants to adjust the data. Default is False.
    k : int, optional
        Degree of the cubic spline. Default is 3.
    s : int, optional
        Smoothing factor. Default is 10e6.
    """

    tank_class: Tank
    freq: str
    position: str
    smooth: Optional[bool]
    s: Optional[float]
    k: Optional[int]

    class Config:
        arbitrary_types_allowed = True

    def __init__(self, tank_class,
                 freq: str,
                 position: str,
                 smooth: bool = False,
                 k: int = 2,
                 s: float = 10e5):

        """
        Parameters
        ----------
        tank_class : Tank
            Instance of the Tank class.
        freq : str
            Frequency of data for the material balance equation.
        position : str
            Position of the frequency of the date.
        smooth : bool, optional
            Determines if the user wants to adjust the data. Default is False.
        k : int, optional
            Degree of the cubic spline. Default is 3.
        s : float, optional
            Smoothing factor. Default is 10e6.
        """
        super().__init__(tank_class=tank_class,
                         freq=freq,
                         position=position,
                         smooth=smooth,
                         k=k,
                         s=s)

    def _calc_uw(self) -> pd.DataFrame:
        """
        Internal method to calculate underground withdrawal (F)
        per well using underground_withdrawal() function.
        :return:
        - pd.Dataframe: A DataFrame with underground withdrawal (F)
         information.
        """

        # Call the internal methods that process production and pressure data
        df_press = self.tank_class.get_pressure_df()
        df_prod = self.tank_class.get_production_df()

        # Validate df_press and df_prod
        df_press_validate = pd.DataFrame(_PressSchema.validate(df_press))
        df_prod_validate = pd.DataFrame(_ProdSchema.validate(df_prod))

        # Calculate the accumulated production in the pressure dataframe,
        # based on the production dataframe.
        for col in [OIL_CUM_COL, WATER_CUM_COL, GAS_CUM_COL]:
            df_press_validate[col] = df_press_validate.apply(
                lambda x: interp_dates_row(
                    x,
                    DATE_COL,
                    df_prod_validate,
                    DATE_COL,
                    col,
                    WELL_COL,
                    WELL_COL,
                    left=0.0,
                ),
                axis=1,
            )
            # For wells not available in the production data frame,
            # fill nans with 0
            df_press_validate.fillna({col: 0.0}, inplace=True)

        uw_well = []  # Empty list to uw values
        for well, group in df_press_validate.groupby(WELL_COL):
            group[UW_COL] = underground_withdrawal(
                group,
                OIL_CUM_COL,
                WATER_CUM_COL,
                GAS_CUM_COL,
                OIL_FVF_COL,
                WATER_FVF_COL,
                GAS_FVF_COL,
                RS_COL,
                RS_W_COL,
            )
            uw_well.append(group)

        df_press_validate = pd.concat(uw_well, ignore_index=True)
        return df_press_validate

    def _pressure_vol_avg(self) -> pd.DataFrame:
        """
        Internal method to calculate the average pressure per tank using
        press_vol_avg() function.
        :return:
            - pd.DataFrame: A DataFrame with the average pressure column
        """
        # Encapsulation of DataFrame from internal private method

        df_press = self._calc_uw()
        df_press_avg = (
            df_press.groupby(TANK_COL).apply(lambda g: pressure_vol_avg(
                g,
                WELL_COL,
                DATE_COL,
                PRESSURE_COL,
                UW_COL,
                self.freq,
                self.position,
            )).reset_index(0))
        if self.smooth is False:
            df_press_avg[PRESSURE_COL] = (df_press_avg[PRESSURE_COL].
                                          interpolate(method="linear"))
        else:
            df_press_avg['DAYS'] = (df_press_avg[DATE_COL] -
                                    df_press_avg[DATE_COL].min()).dt.days
            valid_data = df_press_avg.dropna(subset=[PRESSURE_COL])
            spline = UnivariateSpline(valid_data['DAYS'],
                                      valid_data[PRESSURE_COL],
                                      s=self.s,
                                      k=self.k)
            x_fit = np.linspace(min(df_press_avg["DAYS"]),
                                max(df_press_avg["DAYS"]),
                                len(df_press_avg[DATE_COL]))
            y_fit = spline(x_fit)
            df_press_avg["AVG_PRESS"] = df_press_avg[PRESSURE_COL]
            df_press_avg[PRESSURE_COL] = y_fit

        return df_press_avg

    def mat_bal_df(self) -> pd.DataFrame:
        """
        Obtains the material balance parameters at a certain frequency.

        Returns
        -------
        pandas.DataFrame
            A DataFrame with the following columns:
            - Tank: Name of the tank.
            - START_DATETIME: Date.
            - PRESSURE_DATUM: Pressure value.
            - OIL_CUM_TANK: Oil cumulative production.
            - WATER_CUM_TANK: Water cumulative production.
            - GAS_CUM_TANK: Gas cumulative production.
            - Bo: Oil volumetric factor.
            - Bg: Gas volumetric factor.
            - GOR: Oil Solubility.
            - Bw: Water volumetric factor.
            - Rs_bw: Water Solubility.
            - Time_Step: Time lapses.
            - UW: Underground withdrawal.
            - Eo: Oil Expansion.
            - Eg: Gas Expansion.
            - Efw: Rock-fluid Expansion.
            - Cumulative We: Cumulative influx of water.
        """
        # Encapsulation of DataFrame from internal private method
        avg = self._pressure_vol_avg()

        #  Validate df_prod from _prod_df_int
        prod = pd.DataFrame(
            _ProdSchema.validate(self.tank_class.get_production_df()))

        # Linear interpolated of average pressure

        cols_input = [OIL_CUM_COL, WATER_CUM_COL, GAS_CUM_COL]
        cols_output = ["oil_vol", "water_vol", "gas_vol"]
        prod[cols_output] = ((
            prod.groupby(WELL_COL)[cols_input]).diff().fillna(
            prod[cols_input]))
        cols_group = [DATE_COL, TANK_COL, "oil_vol", "water_vol", "gas_vol"]
        df_tank = (prod[cols_group].groupby(
            cols_group[0:2]).sum().groupby(TANK_COL).cumsum().reset_index())

        # Rename of columns of DataFrame
        df_tank.rename(
            columns={
                "oil_vol": OIL_CUM_COL,
                "water_vol": WATER_CUM_COL,
                "gas_vol": GAS_CUM_COL,
            },
            inplace=True,
        )

        oil_cum_per_tank = OIL_CUM_COL + "_TANK"
        water_cum_per_tank = WATER_CUM_COL + "_TANK"
        gas_cum_per_tank = GAS_CUM_COL + "_TANK"

        # Interpolated Cumulative production
        for col, cum_col in zip(
                [oil_cum_per_tank, water_cum_per_tank, gas_cum_per_tank],
                [OIL_CUM_COL, WATER_CUM_COL, GAS_CUM_COL],
        ):
            avg[col] = avg.apply(
                lambda g: interp_dates_row(g, DATE_COL, df_tank, DATE_COL,
                                           cum_col, TANK_COL, TANK_COL),
                axis=1,
            )

        # Sort by dates of DataFrame
        df_mbal = avg.sort_values(DATE_COL)

        # Interpolated PVT properties from pres_avg
        df_mbal[OIL_FVF_COL] = self.tank_class.oil_model.get_bo_at_press(
            df_mbal[PRESSURE_COL])
        df_mbal[GAS_FVF_COL] = self.tank_class.oil_model.get_bg_at_press(
            df_mbal[PRESSURE_COL])
        df_mbal[RS_COL] = self.tank_class.oil_model.get_rs_at_press(
            df_mbal[PRESSURE_COL])

        # In case properties are calculated using correlations
        if (self.tank_class.water_model.salinity is not None
                and self.tank_class.water_model.temperature is not None
                and self.tank_class.water_model.unit is not None):
            df_mbal[
                WATER_FVF_COL] = self.tank_class.water_model.get_bw_at_press(
                df_mbal[PRESSURE_COL])
            df_mbal[RS_W_COL] = self.tank_class.water_model.get_rs_at_press(
                df_mbal[PRESSURE_COL])

            # In case there are default values for Bw and Rs_w
        else:
            df_mbal[
                WATER_FVF_COL] = self.tank_class.water_model.get_default_bw()
            df_mbal[RS_W_COL] = self.tank_class.water_model.get_default_rs()

        # Creation of time lapses columns
        first_time_lapse = pd.Timedelta(days=pd.to_timedelta(
            df_mbal[DATE_COL].diff().iloc[2], unit="D").days)
        df_mbal["Time_Step"] = first_time_lapse.days
        df_mbal["Time_Step"] = df_mbal["Time_Step"].cumsum()
        df_mbal[GAS_FVF_COL] = df_mbal[GAS_FVF_COL].fillna(0.0)

        # Calculated values of Eo, Eg, Efw and F columns
        mbal_term = ho_terms_equation(
            df_mbal,
            OIL_CUM_TANK,
            WATER_CUM_TANK,
            GAS_CUM_TANK,
            PRESSURE_COL,
            OIL_FVF_COL,
            GAS_FVF_COL,
            RS_COL,
            WATER_FVF_COL,
            RS_W_COL,
            self.tank_class.swo,
            self.tank_class.cw,
            self.tank_class.cf,
            float(self.tank_class.oil_model.get_bo_at_press(
                self.tank_class.pi)),
            float(self.tank_class.oil_model.get_bg_at_press(
                self.tank_class.pi)),
            float(self.tank_class.oil_model.get_bo_at_press(
                self.tank_class.pi)),
            float(self.tank_class.oil_model.get_rs_at_press(
                self.tank_class.pi)),
            self.tank_class.pi,
        )
        mbal_final_per_tank = mbal_term

        # Creation of WE value according to the aquifer model
        if self.tank_class.aquifer is None:
            mbal_final_per_tank[WE] = 0.0

        elif isinstance(self.tank_class.aquifer, Fetkovich):
            # If the aquifer instance is Fetkovich, take the pressure and -
            # time values that Fetkovich needs from mbal_final_per_tank
            pr = list(mbal_final_per_tank[PRESSURE_COL])
            time_step = list(mbal_final_per_tank["Time_Step"])
            self.tank_class.aquifer._set_pr_and_time_step(pr, time_step)
            df = self.tank_class.aquifer.we()
            mbal_final_per_tank = mbal_final_per_tank.join(df["Cumulative We"])

        elif isinstance(self.tank_class.aquifer, CarterTracy):
            # If the aquifer instance is Fetkovich, take the pressure and -
            # time values that Fetkovich needs from mbal_final_per_tank
            pr = list(mbal_final_per_tank[PRESSURE_COL])
            time_step = list(mbal_final_per_tank["Time_Step"])
            self.tank_class.aquifer._set_pr_and_time_step(pr, time_step)
            df = self.tank_class.aquifer.we()
            mbal_final_per_tank = mbal_final_per_tank.join(df["Cumulative We"])

        # final mbal DataFrame
        return mbal_final_per_tank

    # ---------------------- CAMPBELL GRAPH --------------------------------
    def campbell_plot(self,
                      custom_line: bool = False,
                      x1: float = None,
                      y1: float = None,
                      x2: float = None,
                      y2: float = None) -> plt.Figure:
        """
        Method to graphic the Campbell graph to be able to graphically see
        the energy contribution of the aquifer.

        Parameters
        -------
        custom_line (bool) : It can be False if you do not want to draw a line
        with selected points. In case if it can be True
        x1 (float) : The x coordinate of the first point
        y1 (float) : The y coordinate of the first point
        x2 (float) : The x coordinate of the second point
        y2 (float) : The y coordinate of the second point

        Returns
        -------
        plt.Figure: A matplotlib Figure object containing the plot with the
        data selected.
        """
        mbal_df = self.mat_bal_df()
        y = mbal_df[UW_COL] / (mbal_df[OIL_EXP] + mbal_df[RES_EXP])
        x = mbal_df[OIL_CUM_TANK]
        data = pd.DataFrame({"Np": x, "F/Eo+Efw": y})
        fig, ax1 = plt.subplots()
        ax1.scatter(x, y)

        # Graph
        if custom_line is False:
            slope, intercept, r, p, se = stats.linregress(
                data["Np"], data["F/Eo+Efw"])
            reg_line = (slope * data["Np"]) + intercept
            ax1.plot(data["Np"],
                     reg_line,
                     color="green",
                     label="Regression line")

        else:
            slope = (y2 - y1) / (x2 - x1)
            intercept = y1 - slope * x1
            x_values = np.linspace(min(data["Np"]), max(data["Np"]), 100)
            y_values = slope * x_values + intercept
            ax1.plot(x_values, y_values,
                     color="red",
                     label="Custom Line")

        ax1.set_xlabel("Np Cumulative Oil Production [MMStb]")
        ax1.set_ylabel("F/Eo+Efw")
        ax1.set_title("Campbell plot of " +
                      str(self.tank_class.name.
                          replace("_", " ")))
        textstr = (
            "Graph that gives an "
            "\nidea of the energy "
            "\ncontribution of an aquifer"
        )
        props = dict(boxstyle="round", facecolor="grey", alpha=0.5)
        ax1.text(
            0.05,
            0.95,
            textstr,
            transform=ax1.transAxes,
            fontsize=9,
            verticalalignment="top",
            horizontalalignment="left",
            bbox=props,
        )
        ax1.legend(frameon=True, framealpha=0.9, loc="upper right")
        plt.grid(True, linestyle="--", alpha=0.7)

        formattery = FuncFormatter(
            lambda x, pos: "{:.1f}KM".format(x * 1e-9))
        ax1.yaxis.set_major_formatter(formattery)

        formatterx = FuncFormatter(
            lambda x, pos: "{:.1f}M".format(x * 1e-6))
        ax1.xaxis.set_major_formatter(formatterx)
        return fig

    def campbell_data(self) -> pd.DataFrame:
        """
        Method to give the Campbell data..
        :return:
        pd.DataFrame: A pandas DataFrame containing the data with Np and
        F/Eo+Efw columns.
        """
        mbal_df = self.mat_bal_df()
        y = mbal_df[UW_COL] / (mbal_df[OIL_EXP] + mbal_df[RES_EXP])
        x = mbal_df[OIL_CUM_TANK]
        data = pd.DataFrame({"Np": x, "F/Eo+Efw": y})
        return data

    # ------------------ HAVLENA AND ODEH METHOD ----------------------------
    def havlena_odeh_plot(self,
                          custom_line: bool = False,
                          x1: float = None,
                          y1: float = None,
                          x2: float = None,
                          y2: float = None) -> plt.Figure:
        """
        Calculate results based on Havlena and Odeh Methods and show a graphic.

        Parameters
        -------
        custom_line (bool) : It can be False if you do not want to draw a line
        with selected points. In case if it can be True
        x1 (float) : The x coordinate of the first point
        y1 (float) : The y coordinate of the first point
        x2 (float) : The x coordinate of the second point
        y2 (float) : The y coordinate of the second point

        Returns
        -------
        plt.Figure: A graph F - WE vs Et
        """
        name_aquifer = ""
        # Creation of WE value according to the aquifer model
        if self.tank_class.aquifer is None:
            name_aquifer = " without Aquifer"

        elif isinstance(self.tank_class.aquifer, Fetkovich):
            name_aquifer = " with Fetkovich Aquifer Model"

        elif isinstance(self.tank_class.aquifer, CarterTracy):
            name_aquifer = " with Carter Tracy Aquifer Model"

        # Data Processing
        mbal_df = self.mat_bal_df()
        y = mbal_df[UW_COL] - mbal_df[WE]
        x = mbal_df[OIL_EXP] + mbal_df[RES_EXP]
        data = pd.DataFrame({"Eo+Efw": x, "F-We": y})
        fig, ax2 = plt.subplots()
        ax2.scatter(data["Eo+Efw"], data["F-We"], color="blue")

        slope, intercept, r, p, se = stats.linregress(data["Eo+Efw"],
                                                      data["F-We"])

        # Graphic
        # Without points selected
        if custom_line is False:
            reg_line = (slope * data["Eo+Efw"]) + intercept
            ax2.plot(data["Eo+Efw"],
                     reg_line,
                     color="red",
                     label="Regression line")
            # Text in the graph
            textstr = "N [MMStb]: {:.2f}".format(slope / 1000000)
            props = dict(boxstyle="round", facecolor="yellow", alpha=0.5)
            ax2.text(
                0.05,
                0.95,
                textstr,
                transform=ax2.transAxes,
                fontsize=10,
                verticalalignment="top",
                horizontalalignment="left",
                bbox=props,
            )
            ax2.legend(frameon=True, framealpha=0.9, loc="upper right")

        # With points selected
        else:
            slope = (y2 - y1) / (x2 - x1)
            intercept = y1 - slope * x1
            x_values = np.linspace(min(data["Eo+Efw"]), max(data["Eo+Efw"]),
                                   100)
            y_values = slope * x_values + intercept
            ax2.plot(x_values, y_values,
                     color="green",
                     label="Custom Line")
            # Text in the graph
            textstr = "N [MMStb]: {:.2f}".format(slope / 1000000)
            props = dict(boxstyle="round", facecolor="yellow", alpha=0.5)
            ax2.text(
                0.05,
                0.95,
                textstr,
                transform=ax2.transAxes,
                fontsize=10,
                verticalalignment="top",
                horizontalalignment="left",
                bbox=props,
            )
            ax2.legend(frameon=True, framealpha=0.9, loc="upper right")

        ax2.set_xlabel("Eo+Efw")
        ax2.set_ylabel("F-We")
        ax2.set_title("Havlena y Odeh plot of " +
                      str(self.tank_class.name.
                          replace("_", " ")) + name_aquifer)

        plt.grid(True, linestyle="--", alpha=0.7)

        # formatter for the axes in M
        formatter = FuncFormatter(
            lambda x, pos: "{:.2f}M".format(x * 1e-6))
        ax2.yaxis.set_major_formatter(formatter)
        return fig

    def havlena_oded_data(self) -> pd.DataFrame:
        """
        Calculate values based on Havlena and Odeh Methods and show a df.
        :return:
        pd.Dataframe: DataFrame with F-We and Eo+Efw columns.
        """
        mbal_df = self.mat_bal_df()
        y = mbal_df[UW_COL] - mbal_df[WE]
        x = mbal_df[OIL_EXP] + mbal_df[RES_EXP]
        data = pd.DataFrame({"Eo+Efw": x, "F-We": y})
        return data

    # ----------------------- ANALYTIC METHOD -------------------------------
    def analytic_method(self, poes: float,
                        option: str) -> Union[pd.DataFrame, plt.Figure]:
        """
        Method used to calculate the POES through an inferred POES that
        ensures that there is the best match between the behavior of the
        observed pressure and the calculated pressure. the calculation is
        done through calculated_pressure_fetkovich and calculated_pressure_
        carter_tracy functions.

        Parameters
        ----------
        poes : float
            Inferred POES (Petroleum-in-Place) value in MMStb.
        option : str
            Determines the type of result to be returned. Can be either
            "data" or "plot".

        Returns
        -------
        Union[pd.DataFrame, plt.Figure]
            - If option is "data", returns a pandas DataFrame containing the
             Date, Observed Pressure and Calculated Pressure.
            - If option is "plot", returns a matplotlib Figure object
            containing a plot of the observed and calculated pressure over
            time.

        Raises
        ------
        ValueError
            If the option is not "data" or "plot".

        """

        # Encapsulation of material balance DataFrame from mat_bal_df() method
        df = self.mat_bal_df()
        press_calc = []

        # name of aquifer model:
        model_aq_name = ""
        # Fetkovich Aquifer Model
        if isinstance(self.tank_class.aquifer, Fetkovich):
            model_aq_name = "Fetkovich Aquifer Model"
            # Call the function to calculate the new pressure
            press_calc = calculated_pressure_fetkovich(
                df[OIL_CUM_TANK],
                df[WATER_CUM_TANK],
                self.tank_class.cf,
                self.tank_class.water_model.temperature,
                self.tank_class.water_model.salinity,
                self.tank_class.oil_model.data_pvt,
                self.tank_class.aquifer.aq_radius,
                self.tank_class.aquifer.res_radius,
                self.tank_class.aquifer.aq_thickness,
                self.tank_class.aquifer.aq_por,
                self.tank_class.aquifer.theta,
                self.tank_class.aquifer.k,
                self.tank_class.aquifer.water_visc,
                self.tank_class.pi,
                self.tank_class.swo,
                poes,
                PRESSURE_PVT_COL,
                OIL_FVF_COL,
            )

        # Carter Tracy Aquifer Model
        elif isinstance(self.tank_class.aquifer, CarterTracy):
            model_aq_name = "Carter Tracy Aquifer Model"
            press_calc = calculate_pressure_with_carter_tracy(
                df[OIL_CUM_TANK],
                df[WATER_CUM_TANK],
                self.tank_class.cf,
                self.tank_class.water_model.temperature,
                self.tank_class.water_model.salinity,
                self.tank_class.oil_model.data_pvt,
                self.tank_class.aquifer.res_radius,
                self.tank_class.aquifer.aq_thickness,
                self.tank_class.aquifer.aq_por,
                self.tank_class.aquifer.theta,
                self.tank_class.aquifer.aq_perm,
                self.tank_class.aquifer.water_visc,
                df["Time_Step"],
                self.tank_class.pi,
                self.tank_class.swo,
                poes,
                PRESSURE_PVT_COL,
                OIL_FVF_COL,
            )

        # Aad the first date to initial pressure
        dates = df[[DATE_COL, PRESSURE_COL]]
        new_date = df[DATE_COL].min() - pd.Timedelta(days=365)
        n_row = pd.DataFrame({DATE_COL: new_date}, index=[0])
        data = pd.concat([n_row, dates]).reset_index(drop=True)
        data.loc[0, PRESSURE_COL] = self.tank_class.pi

        # Add the Calculated Pressure column
        data["PRESS_CALC"] = press_calc

        if option == "data":
            return data[[DATE_COL, PRESSURE_COL, "PRESS_CALC"]]

        elif option == "plot":
            fig8, ax8 = plt.subplots(figsize=(15, 10))
            ax8.scatter(data[DATE_COL].dt.year,
                        data[PRESSURE_COL],
                        label="Observed Pressure")
            ax8.plot(data[DATE_COL].dt.year,
                     press_calc,
                     c="g",
                     label="Calculated Pressure")
            plt.title(f"Pressure vs Time with {model_aq_name}",
                      fontsize=25)
            plt.xlabel("Time (Years)", fontsize=17)
            plt.ylabel("Pressure (PSI)", fontsize=17)
            ax8.set_ylim(0, 4000)
            plt.yticks(fontsize=15)
            plt.xticks(fontsize=15)
            ax8.grid(axis="both", color="lightgray", linestyle="dashed")
            plt.legend(fontsize=15)
            plt.gcf().autofmt_xdate()
            fig8.suptitle("ANALYTIC METHOD", fontsize=22)
            return fig8

        else:
            raise ValueError("Option no validate. Use 'data' or 'plot'.")

    # The following methods are to do an exploratory data analysis -
    # (EDA) of the Tank:
    # ---------------------- GRAPH SECTION ----------------------------
    def plot_cum_prod_well(self) -> plt.Figure:
        """
        Method to generate a graph.
        :return:
        plt.Figure: Graph of Cumulative Production per Well of Tank.
        """
        # Production Data
        df_prod = self.tank_class.get_production_df()
        df_prod[DATE_COL] = pd.to_datetime(df_prod[DATE_COL])
        df_prod = df_prod.sort_values(by=DATE_COL)
        # Well Group
        df_prod_well = df_prod.groupby(WELL_COL)[[OIL_CUM_COL,
                                                  WATER_CUM_COL]].sum()

        fig, ax = plt.subplots(figsize=(10, 6))
        well_ind = df_prod_well.index
        bar_witd = 0.35
        r1 = range(len(well_ind))
        r2 = [x + bar_witd for x in r1]

        ax.bar(
            r1,
            df_prod_well[OIL_CUM_COL],
            color="black",
            width=bar_witd,
            edgecolor="grey",
            label="Oil Cumulative",
        )
        ax.bar(
            r2,
            df_prod_well[WATER_CUM_COL],
            color="blue",
            width=bar_witd,
            edgecolor="grey",
            label="Water Cumulative",
        )

        ax.set_title(
            "Cumulative Production per Well - " +
            str(self.tank_class.name.replace("_", " ").upper()),
            fontsize=16,
        )
        ax.set_xlabel("Well", fontsize=14)
        ax.set_ylabel("Cumulative Production [Stb]", fontsize=14)
        ax.set_xticks([r + bar_witd / 2 for r in range(len(well_ind))])
        ax.set_xticklabels(
            well_ind,
            rotation=45,
            fontproperties=FontProperties(size=8.5, weight="bold"),
        )
        ax.legend(loc="upper left", fontsize=12)
        plt.grid(True, linestyle="--", alpha=0.7)

        # formatter for the axes in M
        formatter = FuncFormatter(lambda x, pos: "{:.0f}M".format(x * 1e-6))
        ax.yaxis.set_major_formatter(formatter)
        plt.tight_layout()
        return fig

    def plot_flow_rate_well(self):
        """
        Method to generate a graph.
        :return:
        plt.Figure: A graph of Flow Rate vs Time per Rate of Tank.
        """
        # Production Data
        df_prod = self.tank_class.get_production_df()
        df_prod[DATE_COL] = pd.to_datetime(df_prod[DATE_COL])
        df_prod = df_prod.sort_values(by=DATE_COL)

        # days
        # df_prod["Days"] = df_prod[DATE_COL].diff().dt.days.fillna(0)
        # df_prod.loc[df_prod["Days"] == 0, "Days"] = 365

        # Calculate daily production rates from cumulative productions.
        df_prod[OIL_RATE_COL] = df_prod.groupby(
            WELL_COL)[OIL_CUM_COL].diff().fillna(0)
        df_prod[WATER_RATE_COL] = (
            df_prod.groupby(WELL_COL)[WATER_CUM_COL].diff().fillna(0))

        fig, (ax1, ax2) = plt.subplots(nrows=2, figsize=(10, 12), sharex=True)

        wells = df_prod[WELL_COL].unique()
        colors = plt.cm.get_cmap("tab10", len(wells))

        # Plot of Oil Flow Rate
        for i, well in enumerate(wells):
            well_data = df_prod[df_prod[WELL_COL] == well]
            dates = well_data[DATE_COL]

            # Oil rate
            ax1.plot(dates,
                     well_data[OIL_RATE_COL],
                     label=f"{well}",
                     color=colors(i))
            ax1.set_title(
                "Oil Flow Rate vs Time by Well - " +
                str(self.tank_class.name.replace("_", " ").
                    upper()), fontsize=16,
            )
            ax1.set_ylabel("Flow Rate [Stb/year]", fontsize=14)
            ax1.legend(loc="upper left", fontsize=12)
            ax1.grid(True, linestyle="--", alpha=0.7)

            # Water rate
            ax2.plot(dates,
                     well_data[WATER_RATE_COL],
                     label=f"{well}",
                     color=colors(i))
            ax2.set_title(
                "Water Flow Rate vs Time by Well - " +
                str(self.tank_class.name.replace("_", " ").
                    upper()), fontsize=16,
            )
            ax2.set_ylabel("Flow Rate [Stb/year]", fontsize=14)
            ax2.set_xlabel("Date", fontsize=14)
            ax2.legend(loc="upper left", fontsize=12)
            ax2.grid(True, linestyle="--", alpha=0.7)

        fig.autofmt_xdate()
        plt.tight_layout()

        # Set Y-axis limit'
        y_max = max(df_prod[OIL_RATE_COL].max(), df_prod[WATER_RATE_COL].max())
        y_max = (y_max // 20000 + 1) * 20000

        ax1.set_ylim(0, y_max)
        ax1.set_yticks(np.arange(0, y_max + 1, 20000))

        ax2.set_ylim(0, y_max)
        ax2.set_yticks(np.arange(0, y_max + 1, 20000))

        # formatter for the axes in K
        formatter = FuncFormatter(lambda x, pos: "{:.0f}K".format(x * 1e-3))
        ax1.yaxis.set_major_formatter(formatter)

        # formatter for the axes in K
        formatter = FuncFormatter(lambda x, pos: "{:.0f}K".format(x * 1e-3))
        ax2.yaxis.set_major_formatter(formatter)
        return fig

    def plot_cum_prod_time(self) -> plt.Figure:
        """
        Method to generate a graph.
        :return:
        plt.Figure: Graph of Oil and Water Cumulative Production vs Time.
        """
        # Average Pressure Data
        df_press_avg = self.mat_bal_df()
        fig1, ax1 = plt.subplots(figsize=(10, 6))
        colors = ["black", "blue"]
        columns = [OIL_CUM_TANK, WATER_CUM_TANK]

        for i, col in enumerate(columns):
            ax1.plot(df_press_avg[DATE_COL],
                     df_press_avg[col],
                     color=colors[i],
                     label=col)

        ax1.set_title(
            "Cumulative Production per Date - " +
            str(self.tank_class.name.replace("_", " ").upper()),
            fontsize=16,
        )
        ax1.set_xlabel("Date", fontsize=14)
        ax1.set_ylabel("Cumulative Production [MMStb]", fontsize=14)
        ax1.legend(loc="upper left", fontsize=12)

        plt.gcf().autofmt_xdate()
        plt.grid(True, linestyle="--", alpha=0.7)

        # formatter for the axes in M
        formatter = FuncFormatter(lambda x, pos: "{:.1f}M".format(x * 1e-6))
        ax1.yaxis.set_major_formatter(formatter)
        return fig1

    def plot_cum_prod_tot_time(self) -> plt.Figure:
        """
        Method to generate a graph.
        :return:
        plt.Figure: Graph of Total Liquid Cumulative Production vs Time.
        """
        # Average Pressure Data
        df_press_avg = self.mat_bal_df()

        fig2, ax2 = plt.subplots(figsize=(10, 6))
        colors = "skyblue"
        total_liquid = (df_press_avg[OIL_CUM_TANK] +
                        df_press_avg[WATER_CUM_TANK])

        ax2.plot(df_press_avg[DATE_COL],
                 total_liquid,
                 color=colors,
                 label="Total Liquid")

        ax2.set_title(
            "Cumulative Total Liquid Production per Date - " +
            str(self.tank_class.name.replace("_", " ").upper()),
            fontsize=16,
        )
        ax2.set_xlabel("Date", fontsize=14)
        ax2.set_ylabel("Cumulative Production [MMStb]", fontsize=14)
        ax2.legend(loc="upper left", fontsize=12)

        plt.gcf().autofmt_xdate()
        plt.grid(True, linestyle="--", alpha=0.7)

        # formatter for the axes in M
        formatter = FuncFormatter(lambda x, pos: "{:.1f}M".format(x * 1e-6))
        ax2.yaxis.set_major_formatter(formatter)
        return fig2

    def plot_press_time(self) -> plt.Figure:
        """
        Method to generate a graph.
        :return:
        plt.Figure: Graph of normal Pressure vs Time.
        """
        # Pressure Data
        df_press = self.tank_class.get_pressure_df()
        df_press[DATE_COL] = pd.to_datetime(df_press[DATE_COL])
        df_press = df_press.sort_values(by=DATE_COL)

        fig3, ax3 = plt.subplots(figsize=(10, 6))
        color = "green"

        ax3.scatter(df_press[DATE_COL],
                    df_press[PRESSURE_COL],
                    color=color,
                    label="Pressure")

        ax3.set_title(
            "Pressure per Date - " +
            str(df_press[TANK_COL][0].replace("_", " ").upper()),
            fontsize=16,
        )
        ax3.set_xlabel("Date", fontsize=14)
        ax3.set_ylabel("Pressure [PSI]", fontsize=14)
        ax3.legend(loc="upper left", fontsize=12)

        plt.gcf().autofmt_xdate()
        plt.grid(True, linestyle="--", alpha=0.7)
        return fig3

    def plot_press_avg_time(self) -> plt.Figure:
        """
        Method to generate a graph.
        :return:
        plt.Figure: Graph of Average Pressure vs Time.
        """
        # Average Pressure Data
        df_press_avg = self.mat_bal_df()
        fig4, ax4 = plt.subplots(figsize=(10, 6))
        color = "red"
        if self.smooth is True:
            ax4.scatter(
                df_press_avg[DATE_COL],
                df_press_avg["AVG_PRESS"],
                color=color,
                label=" Avg Pressure",
            )
            ax4.plot(
                df_press_avg[DATE_COL],
                df_press_avg[PRESSURE_COL],
                color="blue",
                label="Avg Pressure (Smoothed)",
            )
        else:
            ax4.scatter(
                df_press_avg[DATE_COL],
                df_press_avg[PRESSURE_COL],
                color=color,
                label=" Avg Pressure",
            )

        ax4.set_title(
            "Pressure per Date - " +
            str(self.tank_class.name.replace("_", " ").upper()),
            fontsize=16,
        )
        ax4.set_xlabel("Date", fontsize=14)
        ax4.set_ylabel("Average Pressure[PSI]", fontsize=14)
        ax4.legend(loc="upper left", fontsize=12)

        plt.gcf().autofmt_xdate()
        plt.grid(True, linestyle="--", alpha=0.7)
        return fig4

    def plot_press_liq_cum(self) -> plt.Figure:
        """
        Method to generate a graph.
        :return:
        plt.Figure: Graph of Pressure vs Cumulative Liquids (Oil and Water).
        """
        # Pressure date with Cumulative Production
        df_press_cum = self._calc_uw()
        df_press_cum[DATE_COL] = pd.to_datetime(df_press_cum[DATE_COL])
        df_press_cum = df_press_cum.sort_values(by=PRESSURE_COL)

        fig6, ax6 = plt.subplots(figsize=(10, 6))
        colors = ["black", "blue"]
        columns = [OIL_CUM_COL, WATER_CUM_COL]

        for i, col in enumerate(columns):
            ax6.scatter(
                df_press_cum[PRESSURE_COL],
                df_press_cum[col],
                color=colors[i],
                label=col,
            )

        ax6.set_title(
            "Pressure vs Cumulative Production - " +
            str(self.tank_class.name.replace("_", " ").upper()),
            fontsize=16,
        )
        ax6.set_xlabel("Pressure", fontsize=14)
        ax6.set_ylabel("Cumulative Production", fontsize=14)
        ax6.legend(loc="upper left", fontsize=12)

        plt.gcf().autofmt_xdate()
        plt.grid(True, linestyle="--", alpha=0.7)

        # formatter for the axes in M
        formatter = FuncFormatter(lambda x, pos: "{:.1f}M".format(x * 1e-6))
        ax6.yaxis.set_major_formatter(formatter)
        return fig6

    def plot_press_avg_liq_cum(self) -> plt.Figure:
        """
        Method to generate a graph.
        :return:
        Graph of Average Pressure vs Cumulative Liquids (Oil and Water).
        """
        # Average Pressure Data
        df_press_avg = self.mat_bal_df()
        df_press_avg[DATE_COL] = pd.to_datetime(df_press_avg[DATE_COL])
        df_press_avg = df_press_avg.sort_values(by=PRESSURE_COL)
        fig7, ax7 = plt.subplots(figsize=(10, 6))
        colors = ["black", "blue"]
        columns = [OIL_CUM_TANK, WATER_CUM_TANK]

        for i, col in enumerate(columns):
            ax7.scatter(
                df_press_avg[PRESSURE_COL],
                df_press_avg[col],
                color=colors[i],
                label=col,
            )

        ax7.set_title(
            "Average Pressure vs Cumulative Production - " +
            str(self.tank_class.name.replace("_", " ").upper()),
            fontsize=16,
        )
        ax7.set_xlabel("Average Pressure", fontsize=14)
        ax7.set_ylabel("Cumulative Production", fontsize=14)
        ax7.legend(loc="upper left", fontsize=12)

        plt.gcf().autofmt_xdate()
        plt.grid(True, linestyle="--", alpha=0.7)

        # formatter for the axes in M
        formatter = FuncFormatter(lambda x, pos: "{:.1f}M".format(x * 1e-6))
        ax7.yaxis.set_major_formatter(formatter)
        return fig7

    def plot_flow_rate_tank(self):
        """Method to generate a graph.
        :return:
        plt.Figure: A graph of Flow Rate vs Time by Tank.
        """
        # Production Data
        # Average Pressure Data
        df_prod = self.mat_bal_df()
        df_prod[DATE_COL] = pd.to_datetime(df_prod[DATE_COL])
        df_prod = df_prod.sort_values(by=DATE_COL)

        # Days
        # df_prod["Days"] = df_prod[DATE_COL].diff().dt.days.fillna(0)
        # df_prod.loc[df_prod["Days"] == 0, "Days"] = 365

        # Calculate daily production rates from cumulative productions.
        df_prod[OIL_RATE_COL] = df_prod[OIL_CUM_TANK].diff().fillna(0)
        df_prod[WATER_RATE_COL] = df_prod[WATER_CUM_TANK].diff().fillna(0)

        fig, (ax1, ax2) = plt.subplots(nrows=2, figsize=(10, 12), sharex=True)

        # Oil rate
        ax1.plot(
            df_prod[DATE_COL],
            df_prod[OIL_RATE_COL],
            color="black",
            label="Oil flow Rate",
        )
        ax1.set_title(
            "Oil Flow Rate vs Time by " +
            str(self.tank_class.name.replace("_", " ").upper()),
            fontsize=16,
        )
        ax1.set_ylabel("Flow Rate [Stb/year]", fontsize=14)
        ax1.legend(loc="upper left", fontsize=12)
        ax1.grid(True, linestyle="--", alpha=0.7)

        # Water rate
        ax2.plot(
            df_prod[DATE_COL],
            df_prod[WATER_RATE_COL],
            color="blue",
            label="Water flow rate",
        )
        ax2.set_title(
            "Water Flow Rate vs Time by " +
            str(self.tank_class.name.replace("_", " ").upper()),
            fontsize=16,
        )
        ax2.set_ylabel("Flow Rate [Stb/year]", fontsize=14)
        ax2.set_xlabel("Date", fontsize=14)
        ax2.legend(loc="upper left", fontsize=12)
        ax2.grid(True, linestyle="--", alpha=0.7)

        # formatter for the axes in K
        formatter = FuncFormatter(lambda x, pos: "{:.0f}K".format(x * 1e-3))
        ax1.yaxis.set_major_formatter(formatter)

        # formatter for the axes in K
        formatter = FuncFormatter(lambda x, pos: "{:.0f}K".format(x * 1e-3))
        ax2.yaxis.set_major_formatter(formatter)

        fig.autofmt_xdate()
        plt.tight_layout()

        return fig



"""
utilities.py

This module contains several functions that are helpful for other functions
and classes in the library.

Libraries:
    - calendar
    - datetime
    - pandera
    - pandas
    - numpy
    - typing
"""

import datetime
import pandas as pd
import numpy as np
from pandera import Column, Check, DataFrameSchema
from pytank.constants.constants import DATE_COL, VALID_FREQS, PRESSURE_COL
from typing import Union, Optional, Sequence
from calendar import monthrange


def days_in_month(date):
    if isinstance(date, datetime.datetime):
        return monthrange(date.year, date.month)[1]
    else:
        raise ValueError("Argument is not of type datetime")


def interp_from_dates(date_interp: datetime.datetime,
                      x_dates,
                      y_values,
                      left=None,
                      right=None):
    """
    Interpolation function that accepts dates as x to interpolate between
    y_values, given a pytank date between x_dates

    Parameters
    ----------
    date_interp: array-like of datetime objects
        the pytank date to interpolate
    x_dates: array-like of datetime objects
        the x_dates values to use as regression
    y_values: array-like
        the y_values to use as regression
    left: optional float or complex corresponding to y_values
        Value to return for x < x_dates[0], default is y_values[0]
    right: optional float or complex corresponding to y_values
        Value to return for x > x_dates[-1], default is y_values[-1]
    Returns
    -------
    float or ndarray
        an interpolated value or values between y_values

    """

    # Type checking before proceeding with operations
    permitted_arrays = (list, np.ndarray, pd.Series)
    if isinstance(date_interp, permitted_arrays):
        is_array = True
        if not all(isinstance(x, datetime.datetime) for x in date_interp):
            raise ValueError(
                "date_interp should only contain datetime objects")
    else:
        is_array = False
        if not isinstance(date_interp, datetime.datetime):
            raise ValueError(f"{date_interp} is not a datetime object")

    if isinstance(x_dates, permitted_arrays):
        if not all(isinstance(x, datetime.datetime) for x in x_dates):
            raise ValueError("x_dates should only contain datetime objects")
    else:
        raise ValueError(
            "x_dates should be a list, numpy array or pandas Series")

    if not isinstance(y_values, permitted_arrays):
        raise ValueError(
            "y_values should be a list, numpy array or pandas Series")

    # The handling of numeric values in y_values should be handled by np
    # itself in the interpolation function

    # Get the minimum date in x_dates as the reference
    start_date = x_dates.min()
    # Calculate time deltas using x_dates and the start dates,
    # then convert to seconds
    time_deltas = pd.Series(x_dates - start_date).dt.total_seconds()

    if is_array:
        # do the same for date_interp values
        new_time_delta = pd.Series(date_interp - start_date).dt.total_seconds()
    else:
        new_time_delta = (date_interp - start_date).total_seconds()

    return np.interp(new_time_delta,
                     time_deltas,
                     y_values,
                     left=left,
                     right=right)


def interp_dates_row(
    row,
    x_result_col,
    df_input: pd.DataFrame,
    x_input_col,
    y_input_col,
    input_cond_col_name,
    result_cond_col_name,
    left=None,
    right=None,
):
    """
    A helper function that works on data frame rows using the apply method.
    This function creates an interpolation using interp_from_dates whose
    input values would change based on a condition specified in the target
    data frame

    Parameters
    ----------
    row: DataFrame
        Usually used with apply method, axis=1
    x_result_col: str
        the x value col name to interpolate in the target row
    df_input: DataFrame
        the input data frame that will be used for regression
    x_input_col: str
        the x column in the input data frame
    y_input_col: str
        the y column in the input data frame
    input_cond_col_name: str
        the condition column in the input data frame
    result_cond_col_name: str
        the condition column in the result data frame
    left: optional float or complex corresponding to y_values
        Value to return for x < x_dates[0], default is y_values[0]
    right: optional float or complex corresponding to y_values
        Value to return for x > x_dates[-1], default is y_values[-1]

    Returns
    -------
    float
        returns an interpolated value of float type
    """

    # Filter the input data frame according to group name
    df_input_filt = df_input.loc[df_input[input_cond_col_name] ==
                                 row[result_cond_col_name]]

    if len(df_input_filt) == 0:
        return np.nan
    else:
        return interp_from_dates(
            row[x_result_col],
            df_input_filt[x_input_col],
            df_input_filt[y_input_col],
            left=left,
            right=right,
        )


def material_bal_var_type(data, numb_or_column):
    """
    Function to check the data types of the material balance equation terms

    Parameters
        ----------
    data: Pandas Dataframe
        Contains the production information for a single entity
    numb_or_column: Dictionary
        Python dictionary containing the names of some columns of the
        dataframe in order to check their data types

    Returns
    -------
    Pandas Dataframe:
        Returns the original Pandas Dataframe
    """
    # Make a copy of the original dataframe
    df = data.copy()
    if not isinstance(data, pd.DataFrame):
        raise TypeError("The input data should be a pandas dataframe")

    # Define internal names for column in the DataFrame
    for col, arg in numb_or_column.items():
        if isinstance(arg, (int, float)):
            df[col] = arg
            numb_or_column[col] = col
        elif isinstance(arg, str):
            df.rename(columns={arg: col}, inplace=True)
        else:
            raise TypeError(
                f"{arg} should be either a numeric value or string "
                f"indicating a column in the DataFrame")

    return df


def material_bal_numerical_data(vector):
    """
    Function to check the numerical data types of the arguments of the
    material balance dataframes

    Parameters
    ----------
    vector: List or numpy array
        List or array of numerical arguments for each function
        Contains the production information for a single entity

    Returns
    -------
    A message showing if there is presence of TypeError"""
    for value in vector:
        if not isinstance(value, (int, float)):
            raise TypeError(f"{value} should be either an int or float")


def variable_type(obj):
    """
    :param obj: variable to be converted as an array, it may be a list or
     a float
    :return: an array if obj is entered in the right data format
    """
    if isinstance(obj, np.ndarray):
        array = obj
    elif isinstance(obj, list):
        array = np.array(obj)
    elif isinstance(obj, float):
        array = np.array(obj)
    else:
        raise ValueError(
            "Please enter measured values as float, list or array")
    return array


def add_date_index_validation(base_schema: DataFrameSchema,
                              freq: str = None) -> DataFrameSchema:
    """Add a date index validation to a base schema."""
    if freq is None:
        # iF freq is None, we just add a date column without validation
        new_schema = base_schema.add_columns({
            DATE_COL:
            Column(
                pd.Timestamp,
                coerce=True,
                nullable=False,
                name=None,
            )
        }).set_index([DATE_COL])

        return new_schema

    if freq not in VALID_FREQS:
        raise ValueError(f"freq must be one of {VALID_FREQS}, not {freq}")

    new_schema = base_schema.add_columns({
        DATE_COL:
        Column(
            pd.Timestamp,
            Check(
                lambda s: pd.infer_freq(s) == freq,
                name="DateTimeIndex frequency check",
                error=f"DateTimeIndex must have frequency '{freq}'",
            ),
            coerce=True,
            nullable=False,
            name=None,
        )
    }).set_index([DATE_COL])

    return new_schema


def add_pressure_validation(base_schema: DataFrameSchema) -> DataFrameSchema:
    """Add a pressure column validation to a base schema."""
    new_schema = base_schema.add_columns({
        PRESSURE_COL:
        Column(
            float,
            Check(lambda s: s >= 0),
            coerce=True,
            nullable=False,
            name=None,
            title="Bottom hole pressure",
        )
    })

    return new_schema
    pass


def normalize_date_freq(
    df: Union[pd.DataFrame, pd.Series],
    freq: str,
    start_date: Optional[str] = None,
    end_date: Optional[str] = None,
    cols_fill_na: Optional[Sequence[str]] = None,
    method_no_cols: Optional[str] = None,
    fill_na: Union[int, float] = np.nan,
):
    """
    Returns a pytank dataframe with a pytank DateTimeIndex which has a
    predefined frequency. Using this function will ensure there is always a
    frequency associated to the index of this dataframe.

    Parameters
    ----------
    df: pandas DataFrame or Series
        A pandas DataFrame or Series whose index is of type DateTimeIndex
    freq: str
        A string representing the target frequency for the dataframe.
    start_date
        Specify the start date to reindex the dates. If not specified, the
        earliest date in the index will be used
    end_date
        Specify the end date to reindex the dates. If not specified, the
        latest date in the index will be used
    cols_fill_na
        Data Frame columns to fill with values specified in the argument
        fill_na
    method_no_cols
        {None, ‘backfill’/’bfill’, ‘pad’/’ffill’, ‘nearest’}
        The method to use with other columns not specified in cols_fill_na.
        The options are passed to the pandas.reindex method.
    fill_na: scalar, default np.nan
        Value to use for missing values after the reindexing.

    Returns
    -------
    Series/DataFrame with changed index based on the pytank frequency.
    The pytank dataframe will be sorted by its index.
    """

    if isinstance(df, (pd.DataFrame, pd.Series)):
        if isinstance(df.index, pd.DatetimeIndex):
            # Check for duplicate values in index
            if len(df.index) != len(set(df.index)):
                error_message = (f"The DateTimeIndex contains duplicate "
                                 f"values\n Please, provide dates that are "
                                 f"unique for each production information\n "
                                 f"Printing the first rows of the dataframe:\n"
                                 f" {df.head()}")
                raise IndexError(error_message)
            # First we need to sort the dataframe based on its index
            # to get the start and end dates, just in case.
            sorted_df: pd.DataFrame = df.sort_index()
            start_date_n = (sorted_df.index[0] if start_date is None else
                            pd.to_datetime(start_date))
            end_date_n = (sorted_df.index[-1]
                          if end_date is None else pd.to_datetime(end_date))
        else:
            raise IndexError(
                "The index in the DataFrame or Series should be a "
                "DateTimeIndex object")
    else:
        raise TypeError(
            "First argument should be a pandas DataFrame or Series")

    new_index = pd.date_range(start_date_n,
                              end_date_n,
                              freq=freq,
                              name=df.index.name)
    if cols_fill_na is not None:
        # First reindex the columns that are specified in the cols_fill_na
        # argument
        # These columns will default to nan where pytank dates appear,
        # and will be replaced with the fill_na value
        df_cols = (sorted_df[cols_fill_na].reindex(
            new_index,
            fill_value=fill_na).fillna(method=method_no_cols).reset_index())
        # Reindex the remaining columns, this time by using the method_co_cols
        # argument as input argument to reindex 'method' argument.
        df_no_cols = (
            sorted_df[sorted_df.columns.difference(cols_fill_na)].reindex(
                new_index, method=method_no_cols).reset_index(drop=True))

        df_concat = pd.concat([df_cols, df_no_cols], axis=1)
        df_concat.set_index(df.index.name, inplace=True)
        # Use the same column order as the original dataframe
        df_concat = df_concat[df.columns]
        return df_concat
    else:
        return sorted_df.reindex(new_index, fill_value=fill_na)



# Libraries
import numpy as np


# File to code aquifers-related dataframes or classes
def schilthuis_aq(pi, p, k, h, uw, ra, re):
    """
    This function calculates the aquifers water influx according to schilthuis equation
    :param pi: Initial reservoir pressure (psi)
    :param p: Current reservoir pressure, might be scalar, list or numpy array (psi)
    :param k: effective permeability to water (mili darcy)
    :param h: Estimated reservoir thickness in the aquifers zone (ft)
    :param uw: Water viscosity (cp)
    :param ra: aquifers radius (ft)
    :param re: hydrocarbon zone radius (ft)
    :return: a single or array-like value of aquifers influx (bbl)
    """

    # Convert to numpy array in case the user has input a scalar or list
    # TODO Check only for numerical values and lists, otherwise throw an Exception
    if isinstance(p, np.ndarray):
        p_np = p
    else:
        p_np = np.array(p)

    # Calculate water influx (Field units)
    # TODO handle different units
    c = 0.00708 * k * h / (uw * np.log(ra / re))
    ew = c * (pi - p_np)

    return ew



"""
pvt_interp.py

This module containing a function that allows interpolated some column of
PVT properties

libraries:
    - pandas
    - scipy
"""
import pandas as pd
from scipy import interpolate


def interp_pvt_matbal(pvt: pd.DataFrame, press_col_name: str,
                      prop_col_name: str, press_target: float) -> float:
    """
    Function to calculate PVT properties using lineal interpolate.
    :param:
        - pvt: DataFrame for pvt data
        - press_col_name: String indicating the column name of pressure values
        - prop_col_name: String indicating the column name of pvt property
         values to be interpolated
        - press_target: Numeric value indicating the reservoir pressure target
        value to interpolate to

    :return:
        float: Interpolated pvt property value that match with the
        pressure target
    """
    x = pvt[press_col_name]
    y = pvt[prop_col_name]
    function = interpolate.interp1d(x, y, fill_value='extrapolate')
    return float(function(press_target))



"""
material_balance_functions.py

This module contains the necessary functions to calculate the poes of graphic
and analytical way.

Libraries:
    - pandas
    - numpy
    - matplotlib
    - math
    - scipy
"""

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import math
from scipy import stats
import warnings
from scipy.optimize import fsolve
from pytank.functions.pvt_correlations import Bo_bw, comp_bw_nogas
from pytank.functions.pvt_interp import interp_pvt_matbal
from pytank.functions.utilities import (material_bal_var_type,
                                        material_bal_numerical_data)


# This part of this module contains functions that are used to calculate the
# poes through the graphical method (Havlena and Odeh"


def underground_withdrawal(
        data: pd.DataFrame,
        oil_cum_col: str,
        water_cum_col: str,
        gas_cum_col: str,
        oil_fvf,
        water_fvf,
        gas_fvf,
        gas_oil_rs,
        gas_water_rs,
) -> np.array:
    """
    Calculates the total underground withdrawal of a well using its cumulative
    production information and fluid properties

    Parameters
    ----------
    data: Pandas Dataframe
        Contains the production information for a single entity
    oil_cum_col: str
        Name of oil cumulative column in the data (stb)
    water_cum_col: str
        Name of water cumulative column in the data (stb)
    gas_cum_col: str
        Name of the gas cumulative column in the data (scf)
    oil_fvf: str or float
        Oil formation volume factor column in DataFrame or numeric value
        (rb/stb)
    water_fvf: str or float
        Water formation volume factor in DataFrame or numeric value  (rb/stb)
    gas_fvf: str or float
        Gas formation volume factor in DataFrame or numeric value (rb/scf)
    gas_oil_rs: str or float
        Solution gas-oil ratio in DataFrame or numeric value (scf/stb)
    gas_water_rs: str or float
        Solution gas-water ratio in DataFrame or numeric value (scf/stb)

    Returns
    ------
    Numpy Array:
        Returns numpy array with the total underground withdrawal

    Raises
    ------
    TypeError:
        When the input data is not a pandas DataFrame or the required numeric
        arguments
        are not numeric.

    ArithmeticError:
        If the free gas calculation results in a negative value.

    """

    if not isinstance(data, pd.DataFrame):
        raise TypeError("The input data should be a pandas dataframe")

    df = data.copy()
    # Define internal names for column in the DataFrame
    oil_fvf_col = "oil_fvf"
    water_fvf_col = "water_fvf"
    gas_fvf_col = "gas_fvf"
    rs_col = "gas_oil_rs_col"
    rsw_col = "gas_water_rs"

    # Dictionary containing the names of some columns of the dataframe
    numeric_or_col_args = {
        oil_fvf_col: oil_fvf,
        water_fvf_col: water_fvf,
        gas_fvf_col: gas_fvf,
        rs_col: gas_oil_rs,
        rsw_col: gas_water_rs,
    }

    # Rename the input dataframe and checking the data types of the function
    # arguments
    df = material_bal_var_type(data, numeric_or_col_args)

    # No check for the column types is needed as pandas will raise an error
    # if the key does not exist for the input DataFrame

    # Calculate the incremental oil, water and gas volumes
    oil_vol_col = "oil_vol"
    water_vol_col = "water_vol"
    gas_vol_col = "gas_vol"

    cols_input = [oil_cum_col, water_cum_col, gas_cum_col]
    cols_output = [oil_vol_col, water_vol_col, gas_vol_col]

    df[cols_output] = df[cols_input].diff().fillna(data[cols_input])

    # Calculate gas withdrawal
    gas_withdrawal = (df[gas_vol_col] - df[oil_vol_col] * df[rs_col] -
                      df[water_vol_col] * df[rsw_col]) * df[gas_fvf_col]

    if sum(gas_withdrawal < 0) > 0:
        raise ArithmeticError(
            "Gas withdrawal results in negative values. Consider "
            "adjusting solution gas-oil/water ratio to reflect "
            "consistent gas production")

    gas_withdrawal.fillna(0, inplace=True)

    uw = (df[oil_vol_col] * df[oil_fvf_col] +
          df[water_vol_col] * df[water_fvf_col] + gas_withdrawal)

    return uw.cumsum().values


def pressure_vol_avg(
    data: pd.DataFrame,
    entity_col,
    date_col,
    press_col,
    uw_col,
    avg_freq="1MS",
    position="begin",
) -> pd.DataFrame:
    """
    Parameters
    ----------
    data: pandas DataFrame
        The pressure information containing pressure data, the dates and
        the underground withdrawal for each well. If there are nan values in
        thepressure column these rows will be deleted from the DataFrame.
        If nan values are present in the UW columns, they will be replaced
        by zero. The last case is assuming that the recorded pressures were
        obtained without significant underground withdrawal
    entity_col: str
        The column name where the entities are defined, i.e: wells
    date_col: str
        The column name where the pressure dates are defined
    press_col: str
        The column name where the pressure information is defined
    uw_col: str
        The column name where the underground withdrawal information is defined
    avg_freq: str
        The time frequency at which the pressure volumetric average is required
    position: str
        The position of the grouped date within its interval. Accepted
        values are:
        "begin": dates start at the beginning of the grouped interval
        "middle": dates start in the middle of the grouped interval
        "end": dates start at the end of the grouped interval

    Returns
    -------
    pandas Dataframe
        A DataFrame with the grouped dates and pressure volumetric averages

    Raises
    ------
    ValueError
        If there are underground withdrawal values that are not monotonically
        increasing
    """
    # Avoid warnings
    warnings.filterwarnings(
        "ignore",
        category=FutureWarning,
        message="'M' is deprecated and will be removed in a future version,"
                " please use 'ME' instead.")

    df = data.copy()

    # Make consistency checks
    # Eliminate rows where pressure contain nan values
    df.dropna(subset=[press_col], inplace=True)
    # Replace nan UW values with zeros
    df.fillna(0, inplace=True)
    # Sort by date and check monotonic increase in UW
    df.sort_values(date_col, inplace=True)
    mono_increase = df.groupby(entity_col)[uw_col].is_monotonic_increasing
    for well, mono in mono_increase.items():
        if not mono:
            raise ValueError(
                f"Well {well} contains underground withdrawal values that "
                f"are not increasing with time")

    pos = ["begin", "middle", "end"]

    # Check if position argument has the correct values
    if position not in pos:
        raise ValueError(f"{position} is not an accepted value for 'position' "
                         f"argument. Use any of {pos} instead.")

    # Calculate the differences in pressure and underground withdrawal per well
    delta_uw_col = "delta_uw"
    delta_press_col = "delta_press"

    df[[delta_uw_col, delta_press_col
        ]] = (df.groupby(entity_col)[[uw_col, press_col
                                      ]].diff().fillna(df[[uw_col,
                                                           press_col]]))

    gr_press = df.groupby(pd.Grouper(key=date_col, freq=avg_freq))
    result_avg_press = {date_col: [], press_col: []}

    for group_name, group in gr_press:
        # Identify which values are zero to avoid division by zero
        cond_1 = group[delta_uw_col].abs() > 0
        cond_2 = group[delta_press_col].abs() > 0
        cond = cond_1 & cond_2
        # Assign np.nan in case there are no values
        avg_1 = np.nan
        avg_2 = np.nan

        # This group will be used to get pressure volumetric average as per the
        # defined equation
        g_1 = group[cond]
        if len(g_1) > 0:
            avg_1 = (g_1[press_col] * g_1[delta_uw_col] /
                     g_1[delta_press_col]).sum() / (
                         g_1[delta_uw_col] / g_1[delta_press_col]).sum()

        # This group has no UW and pressure changes, the average of these
        # values will be processed normally
        g_2 = group[~cond]
        if len(g_2) > 0:
            avg_2 = g_2[press_col].mean()

        result_avg_press[date_col].append(group_name)
        # Calculate the average between the volumetric averages and the
        # normal ones
        avg_all = np.array([avg_1, avg_2])
        avg_press = np.nan if all(np.isnan(avg_all)) else np.nanmean(avg_all)

        result_avg_press[press_col].append(avg_press)

    result = pd.DataFrame(result_avg_press)

    if position == pos[0]:
        pass
    else:
        # Get the DateOffset object based on the average frequency
        date_offset = pd.tseries.frequencies.to_offset(avg_freq)
        # Get the information to replicate the date grouping and
        # change accordingly
        start_date = result[date_col].min()
        end_date = result[date_col].max()
        # This date range should be the same as the one generated
        # by the Grouper
        new_dates = pd.date_range(start_date, end_date,
                                  freq=avg_freq) + date_offset
        # Calculate the time deltas comparing to the original dates
        dates_delta = new_dates - result[date_col]

        if position == pos[1]:
            result[date_col] = result[date_col] + dates_delta / 2
        else:
            result[date_col] = result[date_col] + dates_delta

    return result


def oil_expansion(data: pd.DataFrame, oil_fvf, gas_fvf, gas_oil_rs,
                  gas_oil_rs_init, oil_fvf_init) -> pd.Series:
    """
    Calculates the oil expansion using its cumulative production
    information and fluid properties

    Parameters
    ----------
    data: Pandas Dataframe
        Contains the production information for a single entity
    oil_fvf: str or float
        Oil formation volume factor column in DataFrame or numeric value
        (rb/stb)
    gas_fvf: str or float
        Gas formation volume factor column in DataFrame or numeric value
        (rb/scf)
    gas_oil_rs: str or float
        Solution gas-oil ratio column in DataFrame or numeric value (scf/stb)
    gas_oil_rs_init: int or float
        Initial solution gas-oil ratio (scf/stb)
    oil_fvf_init: int or float
        Initial oil formation volume factor (rb/stb)

    Returns
    -------
    Pandas Series:
        Returns Pandas Series with the oil expansion

    Raises
    ------
    TypeError:
        When the input data is not a pandas DataFrame or the required numeric
         arguments
        are not numeric.
    """

    # Define internal names for column in the DataFrame
    oil_fvf_col = "oil_fvf"
    rs_col = "gas_oil_rs "
    gas_fvf_col = "gas_fvf"

    # Dictionary containing the names of some columns of the dataframe
    numeric_or_col_args = {
        oil_fvf_col: oil_fvf,
        rs_col: gas_oil_rs,
        gas_fvf_col: gas_fvf,
    }

    # Rename the input dataframe and checking the data types of the function
    # arguments
    df = material_bal_var_type(data, numeric_or_col_args)

    # Check the data type of numerical arguments
    num_arg = [gas_oil_rs_init, oil_fvf_init]
    material_bal_numerical_data(num_arg)

    tot_fvf_col = df[oil_fvf_col] + (
            (df[rs_col] - gas_oil_rs_init) * df[gas_fvf_col])
    eo = tot_fvf_col - oil_fvf_init
    return eo


# %%
def gas_expansion(data: pd.DataFrame, oil_fvf, gas_fvf, gas_fvf_init,
                  tot_fvf_init) -> pd.Series:
    """
    Calculates the gas expansion using its cumulative production
    information and fluid properties

    Parameters
    ----------
    data: Pandas Dataframe
        Contains the production information for a single entity
    oil_fvf: str or float
        Oil formation volume factor column in DataFrame or numeric value
         (rb/stb)
    gas_fvf: str or float
        Gas formation volume factor in DataFrame or numeric value (rb/scf)
    tot_fvf_init: int or float
        Initial total volume factor of 2 phases (scf/stb)
    gas_fvf_init: int or float
        Initial gas formation volume factor (rb/scf)

     Returns
    -------
    Pandas Series:
        Returns Pandas series with the gas expansion

    Raises
    ------
    TypeError:
        When the input data is not a pandas DataFrame or the required numeric
         arguments
        are not numeric.
    """

    # Define internal names for column in the DataFrame
    oil_fvf_col = "oil_fvf"
    gas_fvf_col = "gas_fvf"

    # Dictionary containing the names of some columns of the dataframe
    numeric_or_col_args = {
        oil_fvf_col: oil_fvf,
        gas_fvf_col: gas_fvf,
    }

    # Rename the input dataframe and checking the data types of the function
    # arguments
    df = material_bal_var_type(data, numeric_or_col_args)

    # Check the data type of numerical arguments
    num_arg = [gas_fvf_init, tot_fvf_init]
    material_bal_numerical_data(num_arg)

    eg = tot_fvf_init * ((df[gas_fvf_col] / gas_fvf_init) - 1)

    return eg


def fw_expansion(
        data: pd.DataFrame,
        oil_fvf,
        p_col: str,
        water_sat,
        water_comp,
        rock_comp,
        oil_fvf_init,
        pressure_init,
) -> pd.Series:
    """
    Calculates the expansion of connate water and rock(formation) using its
    cumulative
    production information and fluid properties

    Parameters
        ----------
    data: Pandas Dataframe
        Contains the production information for a single entity
    oil_fvf: str or float
        Oil formation volume factor column in DataFrame or numeric value
        (rb/stb)
    p_col: str
        Name of the pressure column in the data (psi)
    water_sat: int or float
        Initial water saturation (%)
    water_comp: int or float
        Water compressibility (psi^-1)
    rock_comp: int or float
        Formation (rock) compressibility (psi^-1)
    oil_fvf_init: int or float
        Initial oil formation volume factor (rb/stb)
    pressure_init: int or float
        Initial reservoir pressure (psi)

    Returns
    -------
    Pandas Series:
        Returns Pandas series with the expansion of connate water and
        rock(formation)

    Raises
    ------
    TypeError:
        When the input data is not a pandas DataFrame or the required numeric
        arguments are not numeric.
    """

    # Define internal names for column in the DataFrame
    oil_fvf_col = "oil_fvf"

    # Dictionary containing the oil_fvf column of the dataframe
    numeric_or_col_args = {oil_fvf_col: oil_fvf}

    # Rename the input dataframe and checking the data types of the function
    # arguments
    df = material_bal_var_type(data, numeric_or_col_args)

    # Check the data type of numerical arguments
    num_arg = [water_sat, water_comp, rock_comp, oil_fvf_init, pressure_init]
    material_bal_numerical_data(num_arg)

    efw = (oil_fvf_init * ((water_comp * water_sat + rock_comp) /
                           (1 - water_sat))) * (pressure_init - df[p_col])

    return efw


def ho_terms_equation(
        data: pd.DataFrame,
        oil_cum_col: str,
        water_cum_col: str,
        gas_cum_col: str,
        p_col: str,
        oil_fvf,
        gas_fvf,
        gas_oil_rs,
        water_fvf,
        gas_water_rs,
        water_sat,
        water_comp,
        rock_comp,
        oil_fvf_init,
        gas_fvf_init,
        tot_fvf_init,
        gas_oil_rs_init,
        pressure_init,
) -> pd.DataFrame:
    """
    Calculates the terms of the Havlena and Odeh equation using the cumulative
    production information and fluid properties of some wells and reservoirs

    Parameters
    ----------
    data: Pandas Dataframe
        Contains the production information for a single entity
    oil_cum_col: str
        Name of oil cumulative column in the data (stb)
    water_cum_col: str
        Name of water cumulative column in the data (stb)
    gas_cum_col: str
        Name of the gas cumulative column in the data (scf)
    oil_fvf: str or float
        Oil formation volume factor column in DataFrame or numeric value \
        (rb/stb)
    water_fvf: str or float
        Water formation volume factor in DataFrame or numeric value  (rb/stb)
    gas_fvf: str or float
        Gas formation volume factor in DataFrame or numeric value (rb/scf)
    gas_oil_rs : str or float
        Solution gas-oil ratio in DataFrame or numeric value (scf/stb)
    gas_water_rs : str or float
        Solution gas-water ratio in DataFrame or numeric value (scf/stb)
    p_col: str
        Name of the pressure column in the data (psi)
    water_sat: int or float
        Initial water saturation (%)
    water_comp: int or float
        Water compressibility (psi^-1)
    rock_comp: int or float
        Formation (rock) compressibility (psi^-1)
    oil_fvf_init: int or float
        Initial oil formation volume factor (rb/stb)
    gas_oil_rs_init: int or float
        Initial solution gas-oil ratio (scf/stb)
    gas_fvf_init: int or float
        Initial gas formation volume factor (rb/scf)
    tot_fvf_init: int or float
        Initial total volume factor of 2 phases (scf/stb)
    pressure_init: int or float
        Initial reservoir pressure (psi)

    Returns
    -------
    Pandas Dataframe:
        Returns a pandas concatenated Dataframe with the terms of the Havlena
        and Odeh equation"""

    # Check the data type of numerical arguments
    num_arg = [
        water_sat,
        water_comp,
        rock_comp,
        oil_fvf_init,
        pressure_init,
        gas_oil_rs_init,
        gas_fvf_init,
        tot_fvf_init,
    ]
    material_bal_numerical_data(num_arg)

    # Call the dataframes of the havlena and odeh equation
    f = underground_withdrawal(
        data,
        oil_cum_col,
        water_cum_col,
        gas_cum_col,
        oil_fvf,
        water_fvf,
        gas_fvf,
        gas_oil_rs,
        gas_water_rs,
    )

    eg = gas_expansion(data, oil_fvf, gas_fvf, gas_fvf_init, tot_fvf_init)

    eo = oil_expansion(data, oil_fvf, gas_fvf, gas_oil_rs, gas_oil_rs_init,
                       oil_fvf_init)

    efw = fw_expansion(
        data,
        oil_fvf,
        p_col,
        water_sat,
        water_comp,
        rock_comp,
        oil_fvf_init,
        pressure_init,
    )

    data["UW"] = f
    data["Eo"] = eo
    data["Eg"] = eg
    data["Efw"] = efw

    return data


def campbell_function(
        data: pd.DataFrame,
        oil_cum_col: str,
        water_cum_col: str,
        gas_cum_col: str,
        p_col: str,
        uw_col: str,
        eo_col: str,
        efw_col: str,
        oil_fvf,
        gas_fvf,
        gas_oil_rs,
        water_fvf,
        gas_water_rs,
        water_sat,
        water_comp,
        rock_comp,
        oil_fvf_init,
        gas_fvf_init,
        tot_fvf_init,
        gas_oil_rs_init,
        pressure_init,
):
    """
    This function is able to plot the campbell plot for a required reservoir

    Parameters
    ----------
    data: Pandas Dataframe
        Contains the production information for a single entity
    oil_cum_col: str
        Name of oil cumulative column in the data (stb)
    water_cum_col: str
        Name of water cumulative column in the data (stb)
    gas_cum_col: str
        Name of the gas cumulative column in the data (scf)
    uw_col: str
        Name of the underground withdrawals fluids produced column in the data
    eo_col: str
        Name of the oil expansion column in the data
    efw_col: str
        Name of the column referencing the expansion of the connate water and
         rock in the data
    oil_fvf: str or float
        Oil formation volume factor column in DataFrame or numeric value
        (rb/stb)
    water_fvf: str or float
        Water formation volume factor in DataFrame or numeric value  (rb/stb)
    gas_fvf: str or float
        Gas formation volume factor in DataFrame or numeric value (rb/scf)
    gas_oil_rs : str or float
        Solution gas-oil ratio in DataFrame or numeric value (scf/stb)
    gas_water_rs : str or float
        Solution gas-water ratio in DataFrame or numeric value (scf/stb)
    p_col: str
        Name of the pressure column in the data (psi)
    water_sat: int or float
        Initial water saturation (%)
    water_comp: int or float
        Water compressibility (psi^-1)
    rock_comp: int or float
        Formation (rock) compressibility (psi^-1)
    oil_fvf_init: int or float
        Initial oil formation volume factor (rb/stb)
    tot_fvf_init: int or float
        Initial total volume factor of 2 phases (scf/stb)
    gas_oil_rs_init: int or float
        Initial solution gas-oil ratio (scf/stb)
    gas_fvf_init: int or float
        Initial gas formation volume factor (rb/scf)
    pressure_init: int or float
        Initial reservoir pressure (psi)

    Returns
    -------
    Matplotlib plot:
        Returns a Matplotlib plot of F/Eo+Efw vs  Np (Campbell plot)"""

    # Check the data type of numerical arguments
    num_arg = [
        water_sat,
        water_comp,
        rock_comp,
        oil_fvf_init,
        pressure_init,
        gas_oil_rs_init,
        gas_fvf_init,
        tot_fvf_init,
    ]
    material_bal_numerical_data(num_arg)

    # Call the havlena and odeh terms function
    df = ho_terms_equation(
        data,
        oil_cum_col,
        water_cum_col,
        gas_cum_col,
        p_col,
        oil_fvf,
        gas_fvf,
        gas_oil_rs,
        water_fvf,
        gas_water_rs,
        water_sat,
        water_comp,
        rock_comp,
        oil_fvf_init,
        gas_fvf_init,
        tot_fvf_init,
        gas_oil_rs_init,
        pressure_init,
    )

    vertical_axis = df[uw_col] / (df[eo_col] + df[efw_col])

    # Build the campbell plot
    fig, ax1 = plt.subplots()
    ax1.scatter(df[oil_cum_col], vertical_axis)
    ax1.set_xlabel("Np")
    ax1.set_ylabel("F/Eo+Efw")
    ax1.set_title("Campbell plot")
    plt.show()


def havlena_and_odeh(
        data: pd.DataFrame,
        oil_cum_col: str,
        water_cum_col: str,
        gas_cum_col: str,
        p_col: str,
        uw_col: str,
        eo_col: str,
        eg_col: str,
        oil_fvf,
        gas_fvf,
        gas_oil_rs,
        water_fvf,
        gas_water_rs,
        water_sat,
        water_comp,
        rock_comp,
        oil_fvf_init,
        gas_fvf_init,
        tot_fvf_init,
        gas_oil_rs_init,
        pressure_init,
):
    """
    This function is able to plot the Havlena and Odeh straight line,
    which is useful to determine the OOIP and GIIP of a reservoir. This
    function assumes, that the reservoir contains gas cap and neglect the
    expansion of the connate water and rock

    Parameters
    ----------
    data: Pandas Dataframe
        Contains the production information for a single entity
    oil_cum_col: str
        Name of oil cumulative column in the data (stb)
    water_cum_col: str
        Name of water cumulative column in the data (stb)
    gas_cum_col: str
        Name of the gas cumulative column in the data (scf)
    uw_col: str
        Name of the underground withdrawals fluids produced column in the data
    eo_col: str
        Name of the oil expansion column in the data
    eg_col: str
        Name of the gas expansion column in the data
    oil_fvf: str or float
        Oil formation volume factor column in DataFrame or numeric value
        (rb/stb)
    water_fvf: str or float
        Water formation volume factor in DataFrame or numeric value  (rb/stb)
    gas_fvf: str or float
        Gas formation volume factor in DataFrame or numeric value (rb/scf)
    tot_fvf_init: int or float
        Initial total volume factor of 2 phases (scf/stb)
    gas_oil_rs : str or float
        Solution gas-oil ratio in DataFrame or numeric value (scf/stb)
    gas_water_rs : str or float
        Solution gas-water ratio in DataFrame or numeric value (scf/stb)
    p_col: str
        Name of the pressure column in the data (psi)
    water_sat: int or float
        Initial water saturation (%)
    water_comp: int or float
        Water compressibility (psi^-1)
    rock_comp: int or float
        Formation (rock) compressibility (psi^-1)
    oil_fvf_init: int or float
        Initial oil formation volume factor (rb/stb)
    gas_oil_rs_init: int or float
        Initial solution gas-oil ratio (scf/stb)
    gas_fvf_init: int or float
        Initial gas formation volume factor (rb/scf)
    pressure_init: int or float
        Initial reservoir pressure (psi)

    Returns
    -------
        Returns a Matplotlib plot of F/Eo vs Eg/Eo (Havlena and Odeh Straight
        line)"""

    # Check the data type of numerical arguments
    num_arg = [
        water_sat,
        water_comp,
        rock_comp,
        oil_fvf_init,
        pressure_init,
        gas_oil_rs_init,
        gas_fvf_init,
        tot_fvf_init,
    ]
    material_bal_numerical_data(num_arg)

    # Call the havlena and odeh terms function
    df = ho_terms_equation(
        data,
        oil_cum_col,
        water_cum_col,
        gas_cum_col,
        p_col,
        oil_fvf,
        gas_fvf,
        gas_oil_rs,
        water_fvf,
        gas_water_rs,
        water_sat,
        water_comp,
        rock_comp,
        oil_fvf_init,
        gas_fvf_init,
        tot_fvf_init,
        gas_oil_rs_init,
        pressure_init,
    )

    # Linear regression to calculate the slope and intercept of this
    # straight line
    slope, intercept, r_value, p_value, std_err = stats.linregress(
        df[uw_col] / df[eo_col], df[eg_col] / df[eo_col])

    # Equation of the fitted line using the slope and intercept from
    # linear regression
    y_fit = intercept + (df[eg_col] / df[eo_col] * slope)

    # Build the havlena and odeh straight line
    fig2, ax2 = plt.subplots()
    ax2.plot(
        df[eg_col] / df[eo_col],
        df[uw_col] / df[eo_col],
        marker="o",
        label="original data",
    )
    ax2.plot(df[eg_col] / df[eo_col], y_fit, "r", label="fitted line")
    ax2.set_xlabel("Eg/Eo")
    ax2.set_ylabel("F/Eo")
    ax2.set_title("Havlena and Odeh Straight Line")
    text = "N: %.1f\nmN: %.3f" % (intercept, slope)
    plt.text(0.008, 1, text)
    plt.legend()
    plt.show()


# This part of this module contains functions that are used to calculate the
# poes through the analytical method

def ebm(p: float, pi: float, n_p: float, wp: float, bo: float, cf: float,
        cw: float, sw0: float, boi: float, poes: float, we: float,
        bw) -> float:
    """
    Calculates the function for the Energy Balance Method (EBM).

    Parameters
    ----------
    p : float
        Current reservoir pressure.
    pi : float
        Initial reservoir pressure.
    n_p : float
        Cumulative oil production.
    wp : float
        Cumulative water production.
    bo : float
        Oil formation volume factor.
    cf : float
        Total compressibility.
    cw : float
        Water compressibility.
    sw0 : float
        Initial water saturation.
    boi : float
        Initial oil formation volume factor.
    poes : float
        Inferred Petroleum-in-Place (POES).
    we : float
        Influx of water.
    bw : float
        Water formation volume factor.

    Returns
    -------
    float
        The value of the EBM function.
    """
    eo = bo - boi
    efw = boi * (((cw * sw0) + cf) / (1 - sw0)) * (pi - p)
    f = (n_p * bo) + (wp * bw)
    func_p = (poes * (eo + efw)) + (we * bw) - f
    return func_p


def aquifer_fetkovich(aq_radius: float, res_radius: float, aq_thickness: float,
                      aq_por: float, ct: float, p: float, theta: float,
                      k: float, water_visc: float, last_press: float,
                      cum: float, pi: float) -> float:
    """
    Calculates the accumulated influx of water using a simplified version of
    the Fetkovich class.

    Parameters
    ----------
    aq_radius : float
        Aquifer radius value, ft.
    res_radius : float
        Reservoir radius value, ft.
    aq_thickness : float
        Aquifer thickness, ft.
    aq_por : float
        Aquifer porosity (decimal).
    ct : float
        Total compressibility, psi^-1.
    p : float
        Current reservoir pressure, psi.
    theta : float
        Aquifer angle, degrees.
    k : float
        Permeability value, md.
    water_visc : float
        Viscosity value, cp.
    last_press : float
        Previous reservoir pressure, psi.
    cum : float
        Cumulative influx of water, bbl.
    pi : float
        Initial reservoir pressure, psi.

    Returns
    -------
    float
        Cumulative influx of water, bbl.
    """
    delta_t = 365
    wi = (math.pi / 5.615) * (aq_radius ** 2 -
                              res_radius ** 2) * aq_thickness * aq_por
    f = theta / 360
    wei = ct * wi * pi * f
    rd = aq_radius / res_radius
    j = (0.00708 * k * aq_thickness * f) / (water_visc * (math.log(abs(rd))))
    pa = pi * (1 - (cum / wei))
    pr_avg = (last_press + p) / 2
    we = (wei / pi) * (1 - np.exp(
        (-1 * j * pi * delta_t) / wei)) * (pa - pr_avg)
    cum_water_influx = cum + we
    return cum_water_influx


def fetkovich_press(
        p: float,
        np: float,
        wp: float,
        cf: float,
        t: float,
        salinity: float,
        df_pvt: pd.DataFrame,
        aq_radius: float,
        res_radius: float,
        aq_thickness: float,
        aq_por: float,
        theta: float,
        k: float,
        water_visc: float,
        p_anterior: float,
        cum: float,
        pi: float,
        sw0: float,
        poes: float,
        boi: float,
        ppvt_col: str,
        oil_fvf_col: str,
):
    """
    Calculates the reservoir pressure based on oil properties, oil and water
    production, and aquifer influence.

    Parameters
    ----------
    p : float
        Current reservoir pressure.
    np : float
        Cumulative oil production.
    wp : float
        Cumulative water production.
    cf : float
        Total compressibility.
    t : float
        Temperature.
    salinity : float
        Salinity value.
    df_pvt : pandas.DataFrame
        PVT data frame.
    aq_radius : float
        Aquifer radius.
    res_radius : float
        Reservoir radius.
    aq_thickness : float
        Aquifer thickness.
    aq_por : float
        Aquifer porosity.
    theta : float
        Angle of the aquifer.
    k : float
        Permeability.
    water_visc : float
        Water viscosity.
    p_anterior : float
        Previous reservoir pressure.
    cum : float
        Cumulative influx of water.
    pi : float
        Initial reservoir pressure.
    sw0 : float
        Initial water saturation.
    poes : float
        Inferred POES (Petroleum-in-Place).
    boi : float
        Initial oil formation volume factor.
    ppvt_col : str
        Column name for pressure in the PVT data frame.
    oil_fvf_col : str
        Column name for oil formation volume factor in the PVT data frame.

    Returns
    -------
    float
        The calculated reservoir pressure.
    """
    # Parameters that depend on pressure
    bo = interp_pvt_matbal(df_pvt, ppvt_col, oil_fvf_col, p)
    bw = Bo_bw(p, t, salinity, unit=1)
    cw = comp_bw_nogas(p, t, salinity, unit=1)
    ct = cw + cf
    we = aquifer_fetkovich(
        aq_radius,
        res_radius,
        aq_thickness,
        aq_por,
        ct,
        p,
        theta,
        k,
        water_visc,
        p_anterior,
        cum,
        pi,
    )
    return ebm(p, pi, np, wp, bo, cf, cw, sw0, boi, poes, we, bw)


def calculated_pressure_fetkovich(
        np_frame: pd.Series,
        wp_frame: pd.Series,
        cf: float,
        t: float,
        salinity: float,
        df_pvt: pd.DataFrame,
        aq_radius: float,
        res_radius: float,
        aq_thickness: float,
        aq_por: float,
        theta: float,
        k: float,
        water_visc: float,
        pi: float,
        sw0: float,
        poes: float,
        ppvt_col: str,
        oil_fvf_col: str,
) -> list:
    """
    Calculates the reservoir pressure for each record in the df_ta2
    dataframe using scipy's fsolve function to solve the material balance
    equations iteratively.

    Parameters
    ----------
    np_frame : pandas.Series
        Column of oil cumulative production.
    wp_frame : pandas.Series
        Column of water cumulative production.
    cf : float
        Total compressibility.
    t : float
        Temperature.
    salinity : float
        Salinity value.
    df_pvt : pandas.DataFrame
        PVT data frame.
    aq_radius : float
        Aquifer radius.
    res_radius : float
        Reservoir radius.
    aq_thickness : float
        Aquifer thickness.
    aq_por : float
        Aquifer porosity.
    theta : float
        Angle of the aquifer.
    k : float
        Permeability.
    water_visc : float
        Water viscosity.
    pi : float
        Initial reservoir pressure.
    sw0 : float
        Initial water saturation.
    poes : float
        Inferred POES (Petroleum-in-Place).
    ppvt_col : str
        Column name for pressure in the PVT data frame.
    oil_fvf_col : str
        Column name for oil formation volume factor in the PVT data frame.

    Returns
    -------
    List [float]
        List containing the calculated reservoir pressure.
    """
    # initial values
    boi = interp_pvt_matbal(df_pvt, ppvt_col, oil_fvf_col, pi)
    cum = 0
    calculated_p = [pi]
    x0 = pi

    # Iteration of each of the years
    for i in range(len(np_frame)):
        np = np_frame[i]
        wp = wp_frame[i]
        p_anterior = calculated_p[i]
        # Calculate current reservoir pressure given all other material
        # balance variables through numeric solving.
        pressure = float(
            fsolve(
                fetkovich_press,
                x0,
                args=(
                    np,
                    wp,
                    cf,
                    t,
                    salinity,
                    df_pvt,
                    aq_radius,
                    res_radius,
                    aq_thickness,
                    aq_por,
                    theta,
                    k,
                    water_visc,
                    p_anterior,
                    cum,
                    pi,
                    sw0,
                    poes,
                    boi,
                    ppvt_col,
                    oil_fvf_col,
                ),
            )[0])
        x0 = pressure
        calculated_p.append(pressure)
        cw = comp_bw_nogas(pressure, t, salinity, unit=1)
        ct = cf + cw
        cum = aquifer_fetkovich(
            aq_radius,
            res_radius,
            aq_thickness,
            aq_por,
            ct,
            pressure,
            theta,
            k,
            water_visc,
            p_anterior,
            cum,
            pi,
        )
    return calculated_p


def aquifer_carter_tracy(
        aq_por: float,
        ct: float,
        res_radius: float,
        aq_thickness: float,
        theta: float,
        k: float,
        water_visc: float,
        pr: float,
        time: float,
        past_time: float,
        we: float,
        pi: float,
) -> float:
    """
    Calculates the accumulated influx of water using a simplified version of
    the Carter-Tracy class.

    Parameters
    ----------
    aq_por : float
        Aquifer porosity (decimal).
    ct : float
        Total compressibility, psi^-1.
    res_radius : float
        Reservoir radius, ft.
    aq_thickness : float
        Aquifer thickness, ft.
    theta : float
        Aquifer angle, degrees.
    k : float
        Permeability, md.
    water_visc : float
        Water viscosity, cp.
    pr : float
        Current reservoir pressure, psi.
    time : float
        Current time.
    past_time : float
        Previous time.
    we : float
        Cumulative water influx, bbl.
    pi : float
        Initial reservoir pressure, psi.

    Returns
    -------
    float
        Cumulative influx of water, bbl.
    """
    pr_array = pr

    # Calculate the van Everdingen-Hurst water influx constant
    f = theta / 360
    b = 1.119 * aq_por * ct * (res_radius ** 2) * aq_thickness * f

    # Estimate dimensionless time (tD)
    cte = 0.006328 * k / (aq_por * water_visc * ct * (res_radius ** 2))
    td = time * cte
    td2 = past_time * cte
    # Calculate the total pressure drop (Pi-Pn) as an array, for each
    # time step n.
    pr_drop = pi - pr_array
    # Estimate the dimensionless pressure
    pr_d = 0.5 * (np.log(td) + 0.80907)
    # Estimate the dimensionless pressure derivative
    """e = 716.441 + (46.7984 * (td**0.5)) + (270.038 * td) + (71.0098 *
                                                            (td**1.5))
    d = ((1296.86 * (td**0.5)) + (1204.73 * td) + (618.618 * (td**1.5)) +
         (538.072 * (td**2)) + (142.41 * (td**2.5)))"""

    pr_deriv = 1 / (2 * td)

    a1 = td - td2
    a2 = b * pr_drop
    a3 = we * pr_deriv
    a4 = pr_d
    a5 = td2 * pr_deriv
    cum_influx_water = we + (a1 * ((a2 - a3) / (a4 - a5)))
    we = cum_influx_water
    return we


def carter_tracy_press(p: float, np: float, wp: float, cf: float, t: float,
                       salinity: float, df_pvt: pd.DataFrame,
                       res_radius: float, aq_thickness: float, aq_por: float,
                       theta: float, k: float, water_visc: float, time: float,
                       past_time: float, we: float, pi: float, sw0: float,
                       poes: float, boi: float, ppvt_col: str,
                       oil_fvf_col: str):
    """
    Calculates the reservoir pressure based on oil properties, oil and water
    production, and aquifer influence.

    Parameters
    ----------
    p : float
        Current reservoir pressure.
    np : float
        Cumulative oil production.
    wp : float
        Cumulative water production.
    cf : float
        Formation compressibility.
    t : float
        Temperature.
    salinity : float
        Salinity.
    df_pvt : pandas.DataFrame
        PVT data frame.
    res_radius : float
        Reservoir radius.
    aq_thickness : float
        Aquifer thickness.
    aq_por : float
        Aquifer porosity.
    theta : float
        Aquifer angle in degrees.
    k : float
        Permeability.
    water_visc : float
        Water viscosity.
    time : float
        Current time.
    past_time : float
        Previous time.
    we : float
        Cumulative water influx.
    pi : float
        Initial pressure.
    sw0 : float
        Initial water saturation.
    poes : float
        Estimated Original Oil in Place (OOIP).
    boi : float
        Initial oil formation volume factor.
    ppvt_col : str
        Column name for pressure in the PVT data frame.
    oil_fvf_col : str
        Column name for oil formation volume factor in the PVT data frame.

    Returns
    -------
    float
        The calculated reservoir pressure.
    """
    bo = interp_pvt_matbal(df_pvt, ppvt_col, oil_fvf_col, p)

    bw = Bo_bw(p, t, salinity, unit=1)
    cw = comp_bw_nogas(p, t, salinity, unit=1)
    ct = cw + cf
    we = aquifer_carter_tracy(
        aq_por,
        ct,
        res_radius,
        aq_thickness,
        theta,
        k,
        water_visc,
        p,
        time,
        past_time,
        we,
        pi,
    )
    return ebm(p, pi, np, wp, bo, cf, cw, sw0, boi, poes, we, bw)


def calculate_pressure_with_carter_tracy(np_frame: pd.Series,
                                         wp_frame: pd.Series, cf: float,
                                         t: float, salinity: float,
                                         df_pvt: pd.DataFrame,
                                         res_radius: float,
                                         aq_thickness: float, aq_por: float,
                                         theta: float, k: float,
                                         water_visc: float,
                                         time_frame: pd.Series, pi: float,
                                         sw0: float, poes: float,
                                         ppvt_col: str,
                                         oil_fvf_col: str) -> list:
    """
    Calculates the reservoir pressure for each record in the df_ta2
    dataframe using scipy's fsolve function to solve the material balance
    equations iteratively.

    Parameters
    ----------
    np_frame : pandas.Series
        Column of oil cumulative production.
    wp_frame : pandas.Series
        Column of water cumulative production.
    cf : float
        Total compressibility.
    t : float
        Temperature.
    salinity : float
        Salinity value.
    df_pvt : pandas.DataFrame
        PVT data frame.
    res_radius : float
        Reservoir radius.
    aq_thickness : float
        Aquifer thickness.
    aq_por : float
        Aquifer porosity.
    theta : float
        Angle of the aquifer.
    k : float
        Permeability.
    water_visc : float
        Water viscosity.
    time_frame : pandas.Series
        Column of time steps.
    pi : float
        Initial reservoir pressure.
    sw0 : float
        Initial water saturation.
    poes : float
        Inferred original petroleum in situ.
    ppvt_col : str
        Column name for pressure in the PVT data frame.
    oil_fvf_col : str
        Column name for oil formation volume factor in the PVT data frame.

    Returns
    -------
    List[float]
        List containing the calculated reservoir pressure.
    """
    boi = interp_pvt_matbal(df_pvt, ppvt_col, oil_fvf_col, pi)
    cum = 0
    x0 = pi
    calculated_p = [pi]
    past_time = 0
    for i in range(len(np_frame)):
        np = np_frame[i]
        wp = wp_frame[i]
        time = time_frame[i]
        pressure = float(
            fsolve(
                carter_tracy_press,
                x0,
                args=(np, wp, cf, t, salinity, df_pvt, res_radius,
                      aq_thickness, aq_por, theta, k, water_visc, time,
                      past_time, cum, pi, sw0, poes, boi, ppvt_col,
                      oil_fvf_col),
            )[0])
        x0 = pressure
        calculated_p.append(pressure)
        cw = comp_bw_nogas(pressure, t, salinity, unit=1)
        ct = cf + cw
        cum = aquifer_carter_tracy(
            aq_por,
            ct,
            res_radius,
            aq_thickness,
            theta,
            k,
            water_visc,
            pressure,
            time,
            past_time,
            cum,
            pi,
        )
        past_time = time_frame[i]

    return calculated_p
